from .common import *


@ti.data_oriented
class Scene:
    def __init__(self, res=512, **options):
        '''
        :param res: (int | tuple) resolution of screen
        :param options: options for the rasterizers

        Creates a Tina scene, the top level structure to manage everything in your scene.
        '''

        self.engine = tina.Engine(res)
        self.res = self.engine.res
        self.options = options

        self.image = ti.Vector.field(3, float, self.res)
        self.lighting = tina.Lighting()
        self.default_material = tina.Lambert()
        self.shaders = {}
        self.objects = {}

        self.taa = options.get('taa', False)
        if self.taa:
            self.accum = tina.Accumator(self.res)

        @ti.materialize_callback
        def init_light():
            self.lighting.add_light(dir=[0, 1, 1], color=[0.9, 0.9, 0.9])
            self.lighting.set_ambient_light([0.1, 0.1, 0.1])

    def _ensure_material_shader(self, material):
        if material not in self.shaders:
            shader = tina.Shader(self.image, self.lighting, material)
            self.shaders[material] = shader

    def add_object(self, object, material=None, raster=None):
        '''
        :param object: (Mesh | Pars | Voxl) object to add into the scene
        :param material: (Material) specify material for shading the object, self.default_material by default
        :param raster: (Rasterizer) specify the rasterizer for this object, automatically guess if not specified
        '''

        assert object not in self.objects
        if material is None:
            material = self.default_material

        if raster is None:
            if hasattr(object, 'get_nfaces'):
                if not hasattr(self, 'triangle_raster'):
                    self.triangle_raster = tina.TriangleRaster(self.engine, **self.options)
                raster = self.triangle_raster
            elif hasattr(object, 'get_npars'):
                if not hasattr(self, 'particle_raster'):
                    self.particle_raster = tina.ParticleRaster(self.engine, **self.options)
                raster = self.particle_raster
            elif hasattr(object, 'sample_volume'):
                if not hasattr(self, 'volume_raster'):
                    self.volume_raster = tina.VolumeRaster(self.engine, **self.options)
                raster = self.volume_raster
            else:
                raise ValueError(f'cannot determine raster type of object: {object}')

        self._ensure_material_shader(material)

        self.objects[object] = namespace(material=material, raster=raster)

    def init_control(self, gui, center=None, theta=None, phi=None, radius=None):
        '''
        :param gui: (GUI) the GUI to bind with
        :param center: (3 * [float]) the target (lookat) position
        :param theta: (float) the altitude of camera
        :param phi: (float) the longitude of camera
        :param radius: (float) the distance from camera to target
        '''

        self.control = tina.Control(gui)
        if center is not None:
            self.control.center[:] = center
        if theta is not None:
            self.control.theta = theta
        if phi is not None:
            self.control.phi = phi
        if radius is not None:
            self.control.radius = radius

    def render(self):
        '''
        Render the image to field self.img
        '''

        if self.taa:
            self.engine.randomize_bias(self.accum.count[None] == 0)

        self.image.fill(0)
        self.engine.clear_depth()

        for object, o in self.objects.items():
            shader = self.shaders[o.material]
            o.raster.set_object(object)
            o.raster.render(shader)

        if self.taa:
            self.accum.update(self.image)

    @property
    def img(self):
        '''
        The image to be displayed in GUI
        '''

        if self.taa:
            return self.accum.img
        return self.image

    def input(self, gui):
        '''
        :param gui: (GUI) GUI to recieve event from

        Feed inputs from the mouse drag events on GUI to control the camera
        '''

        if not hasattr(self, 'control'):
            self.control = tina.Control(gui)
        changed = self.control.get_camera(self.engine)
        if changed and self.taa:
            self.accum.clear()
        return changed
